#include "CDetector.hpp"
#include <cfloat>
#include <math.h>
#include <cstdio>

#define PI 3.1415
#define FILTER_EDGES

bool AreOverlapping(RBody* A, RBody* B, Vector3 axis)
{
  double min_a, min_b, max_a, max_b;
  Vector3 va, vb;
  
  va = A->PointLocalToGlobal(A->vertices[0]);
  vb = B->PointLocalToGlobal(B->vertices[0]);
  min_a = max_a = va*axis;
  min_b = max_b = vb*axis;
  
  for (unsigned int i = 1; i < A->vertices.size(); i++)
  {
    va = A->PointLocalToGlobal(A->vertices[i]);
    double val = va*axis;
    if (val < min_a) min_a = val;
    if (val > max_a) max_a = val;
  }
  for (unsigned int i = 1; i < B->vertices.size(); i++)
  {
    vb = B->PointLocalToGlobal(B->vertices[i]);
    double val = vb*axis;
    if (val < min_b) min_b = val;
    if (val > max_b) max_b = val;
  }
  
    // min_b inside A
  if ((min_b >= min_a && min_b <= max_a) || (max_b >= min_a && max_b <= max_a) ||
      (min_a >= min_b && min_a <= max_b) || (max_a >= min_b && max_a <= max_b))
    return true;
  else
    return false;
}

/*double OverlapDepth(RBody* A, RBody* B, Vector3 axis)
{
  double min_a, min_b, max_a, max_b;
  Vector3 va, vb;
  
  va = A->PointLocalToGlobal(A->vertices[0]);
  vb = B->PointLocalToGlobal(B->vertices[0]);
  min_a = max_a = va*axis;
  min_b = max_b = vb*axis;
  
  for (unsigned int i = 1; i < A->vertices.size(); i++)
  {
    va = A->PointLocalToGlobal(A->vertices[i]);
    double val = va*axis;
    if (val < min_a) min_a = val;
    if (val > max_a) max_a = val;
  }
  for (unsigned int i = 1; i < B->vertices.size(); i++)
  {
    vb = B->PointLocalToGlobal(B->vertices[i]);
    double val = vb*axis;
    if (val < min_b) min_b = val;
    if (val > max_b) max_b = val;
  }
  
  double max_depth = -1;
  //double current_depth;
  
  max_depth = max_b - min_a;
    
  return max_depth;
}*/

double OverlapDepth(RBody* A, RBody* B, Vector3 axis, Vector3* midpoint = NULL)
{
  double min_a, min_b, max_a, max_b;
  Vector3 min_av, min_bv, max_av, max_bv;
  Vector3 va, vb;
  
  va = A->PointLocalToGlobal(A->vertices[0]);
  vb = B->PointLocalToGlobal(B->vertices[0]);
  min_a = max_a = va*axis;
  min_b = max_b = vb*axis;
  min_av = max_av = va;
  min_bv = max_bv = vb;
  
  for (unsigned int i = 1; i < A->vertices.size(); i++)
  {
    va = A->PointLocalToGlobal(A->vertices[i]);
    double val = va*axis;
    if (val < min_a) 
    {
      min_a = val;
      min_av = va;
    }
    if (val > max_a)
    {
      max_a = val;
      max_av = va;
    }
  }
  for (unsigned int i = 1; i < B->vertices.size(); i++)
  {
    vb = B->PointLocalToGlobal(B->vertices[i]);
    double val = vb*axis;
    if (val < min_b)
    {
      min_b = val;
      min_bv = vb;
    }
    if (val > max_b)
    {
      max_b = val;
      max_bv = vb;
    }
  }
  
  if (midpoint != NULL)
    *midpoint = (max_bv + min_av)/2;
  
  double max_depth = max_b - min_a;    
  return max_depth;
}

Face* MinimumPenFaceSecond(RBody* A, RBody* B)
{
  double min_pen = DBL_MAX;
  Face* min_face = &(B->faces[0]);
  
  for (int i = 0; i < B->FaceNum(); i++)
  {
    Vector3 normal = B->RotationMatrix()*B->GetFace(i).n;
    double pen = OverlapDepth(A, B, normal);
    if (pen < min_pen) 
    {
      min_pen = pen;
      min_face = &(B->faces[i]);
    }
  }
  return min_face;
}

double CDetector::LineDistance(Edge A, Edge B)
{
  Vector3 w0 = A.a - B.a;
  double a = A.dir * A.dir;
  double b = A.dir * B.dir;
  double c = B.dir * B.dir;
  double d = A.dir * w0;
  double e = B.dir * w0;
  
  Vector3 wc = (w0 + (A.dir*(b*e - c*d) - B.dir*(a*e - b*d))/(a*c - b*b));
  return wc.Length();
}

Vector3 CDetector::ClosestOnEdgeA(Edge A, Edge B)
{
  Vector3 U = A.b-A.a;
  Vector3 V = B.b-B.a;
  Vector3 W = U^V;
  
  // if edges perpendicular return edge end
  if (W.Length() < 0.001)
    return A.a;
  
  double pos_A = (((B.a-A.a)^V)*W)/(W*W);
  //double pos_B = (((B.a-A.a)^U)*W)/(W*W);
  Vector3 cA;
  
  if (pos_A <= 0) cA = A.a;
  else if (pos_A >= 1) cA = A.b;
  else cA = A.a + U*pos_A;
  
  //Vector3 Y = B.a + V*pos_B;
  return cA;
}

bool CDetector::DetectCollision(RBody* A, RBody* B)
{
  double dist = (A->pos - B->pos).Length();
  if (dist > A->bounding_radius + B->bounding_radius)
    return false;
  
  // check last collision points of A
  for (int i = 0; i < A->recent_cols_size; i++)
  {
    if (IsPointInside(A->PointLocalToGlobal(A->recent_cols[i]), B))
      return true;
  }
  
  // check last collision points of B
  for (int i = 0; i < B->recent_cols_size; i++)
  {
    if (IsPointInside(B->PointLocalToGlobal(B->recent_cols[i]), A))
      return true;
  }
  
  // project on axis perpendicular to face of A
  for (int i = 0; i < A->FaceNum(); i++)
  {
    Vector3 normal = A->RotationMatrix()*A->GetFace(i).n*-1;
    if (!AreOverlapping(A, B, normal)) return false;
  }
  
  // project on axis perpendicular to face of B
  for (int i = 0; i < B->FaceNum(); i++)
  {
    Vector3 normal = B->RotationMatrix()*B->GetFace(i).n;
    if (!AreOverlapping(A, B, normal)) return false;
  }
  
  // project on axis along the cross product of edges
  for (unsigned int i = 0; i < A->edges.size(); i++)
    for (unsigned int j = 0; j < B->edges.size(); j++)
    {
      Vector3 da = A->RotationMatrix()*(A->edges[i].dir);
      Vector3 db = B->RotationMatrix()*(B->edges[j].dir);
      Vector3 normal = (da^db).Normalized();
      if (!AreOverlapping(A, B, normal)) return false;
    }
  return true;
}

Vector3 SeparatingAxis(RBody* A, RBody* B)
{
  Vector3 min_normal; // pointing from B to A
  double min_depth = DBL_MAX;
  
  // project on axis perpendicular to face of A
  for (int i = 0; i < A->FaceNum(); i++)
  {
    Vector3 normal = A->RotationMatrix()*A->GetFace(i).n*-1;
    double depth = OverlapDepth(A, B, normal);
    if (depth < min_depth)
    {
      min_depth = depth;
      min_normal = normal;
    }
  }
  
  // project on axis perpendicular to face of B
  for (int i = 0; i < B->FaceNum(); i++)
  {
    Vector3 normal = B->RotationMatrix()*B->GetFace(i).n;
    double depth = OverlapDepth(A, B, normal);
    if (depth < min_depth)
    {
      min_depth = depth;
      min_normal = normal;
    }
  }
  
  // project on axis along the cross product of edges
  for (unsigned int i = 0; i < A->edges.size(); i++)
    for (unsigned int j = 0; j < B->edges.size(); j++)
    {
      Vector3 da = A->RotationMatrix()*(A->edges[i].dir);
      Vector3 db = B->RotationMatrix()*(B->edges[j].dir);
      
      if ((da^db).Length() < 0.001) continue; // edges parallel
      
      Vector3 normal = (da^db).Normalized();
      
      double depth = OverlapDepth(A, B, normal);
      double depth_inv = OverlapDepth(A, B, normal*-1);
      
      if (depth < min_depth)
      {
        min_depth = depth;
        min_normal = normal;
      }
      if (depth_inv < min_depth)
      {
        min_depth = depth_inv;
        min_normal = normal*-1;
      }
    }
    
  return min_normal;
}

bool CDetector::IsPointInside(Vector3 p, RBody* B)
{
   // project on axis perpendicular to face of B
  for (int i = 0; i < B->FaceNum(); i++)
  {
    Vector3 normal = B->RotationMatrix()*B->GetFace(i).n;
    Vector3 to_point = p - (B->RotationMatrix()*B->GetFace(i).vertices[0] + B->pos);
    double pen = -(to_point * normal);
    if (pen < 0) 
    {
      return false;
    }
  }
  return true;
}

bool IsPointAlmostInside(Vector3 p, RBody* B)
{
   // project on axis perpendicular to face of B
  for (int i = 0; i < B->FaceNum(); i++)
  {
    Vector3 normal = B->RotationMatrix()*B->GetFace(i).n;
    Vector3 to_point = p - (B->RotationMatrix()*B->GetFace(i).vertices[0] + B->pos);
    double pen = -(to_point * normal);
    if (pen < -0.1) 
    {
      return false;
    }
  }
  return true;
}

// cp belongs to A
PointEvaluation CDetector::EvaluatePoint(CollisionPoint cp, RBody* A, RBody* B)
{
  PointEvaluation ep;
  ep.p = A->PointLocalToGlobal(cp.p);
  ep.depth = DBL_MAX;
  
  if (cp.eA != NULL && cp.eB != NULL)
  {
    /*Edge ea = A->RotationMatrix() * (*(cp.eA)) + A->pos;
    Edge eb = B->RotationMatrix() * (*(cp.eB)) + B->pos;
    
    Vector3 ca = ClosestOnEdgeA(ea, eb);
    Vector3 cb = ClosestOnEdgeA(eb, ea);
    
    ep.depth = (cb - ca).Length();
    ep.n = (ea.dir ^ eb.dir).Normalized();
    double normal_depth = fabs(OverlapDepth(A, B, ep.n));
    if (normal_depth > ep.depth + 0.01)
      ep.n = ep.n * -1;
    */  
    ep.n = cp.n;
    ep.depth = OverlapDepth(A, B, ep.n);
    //ep.depth = -((ca - cb)*ep.n);
    return ep;
  }
  else
  {
    /*Face* face = cp.f;
    Vector3 normal;
    double pen;
    normal = B->RotationMatrix()*(face->n);
    Vector3 to_point = ep.p - (B->RotationMatrix()*(face->vertices[0]) + B->pos);
    pen = -(to_point * normal);
    ep.depth = pen;
    ep.n = normal;*/
    
    ep.n = cp.n;
    double depth1 = OverlapDepth(A, B, ep.n);
    double depth2 = OverlapDepth(B, A, ep.n);
    ep.depth = std::min(depth1, depth2);
    
  }
  
  return ep;
}

double CDetector::RelativeVelocity(Vector3 p, Vector3 n, RBody* A, RBody* B)
{
  Vector3 pa = p - A->pos;
  Vector3 pb = p - B->pos;
  Vector3 vpa = A->vel + (A->ang_vel ^ pa);
  Vector3 vpb = B->vel + (B->ang_vel ^ pb);
  return n*(vpa - vpb);
}

double HighestRelVel(Collision* col)
{
  double max_rel_vel = 0;
  
  for (int body = 0; body < 2; body++)
  {
    for (unsigned int i = 0; i < col->points[body].size(); i++)
    {
      CollisionPoint cp = col->points[body][i];
      PointEvaluation pe = CDetector::EvaluatePoint(cp, col->bodies[body], col->bodies[(body+1)%2]);
      double rel_vel = CDetector::RelativeVelocity(pe.p, pe.n, col->bodies[body], col->bodies[(body+1)%2]);
      if (fabs(rel_vel) > fabs(max_rel_vel))
        max_rel_vel = rel_vel;
    }
  }
  return max_rel_vel;
}

/*void CDetector::RecalculatePoints(Collision* col)
{
  for (unsigned int i = 1; i < col->points.size(); i++)
  {
    col->points[i].p = col->points[i].pointer->PointLocalToGlobal(col->points[i].pL);
    col->points[i].n = col->points[i].normaler->VectorLocalToGlobal(col->points[i].nL);
    col->points[i].depth = CDetector::EvaluatePoint(col->points[i].p, col->points[i].normaler).depth;
    col->points[i].rel_vel = 
      CDetector::RelativeVelocity(col->points[i].p, col->points[i].n, col->A, col->B);
  }
}*/

Collision* CDetector::GetCollisionPoints(RBody* A, RBody* B)
{
  Collision* ret = new Collision();
  ret->bodies[0] = A;
  ret->bodies[1] = B;
  
  Vector3 normal_b_to_a = SeparatingAxis(A, B);
  Vector3 separating_plane_point;
  double depth = OverlapDepth(A, B, normal_b_to_a, &separating_plane_point);

  std::vector<Vector3> verticesA;
  std::vector<Vector3> verticesB;

#ifdef FILTER_EDGES
  Vector3 plane_point_local_A = A->PointGlobalToLocal(separating_plane_point);
  Vector3 plane_point_local_B = B->PointGlobalToLocal(separating_plane_point);
  Vector3 normal_local_A = A->VectorGlobalToLocal(normal_b_to_a);
  Vector3 normal_local_B = B->VectorGlobalToLocal(normal_b_to_a);
  
  for (unsigned int i = 0; i < A->vertices.size(); i++)
  {
    Vector3 v = A->vertices[i];
    if (abs((v - plane_point_local_A)*normal_local_A) < depth+0.1)
    {
      verticesA.push_back(A->vertices[i]);
    }
  }
  for (unsigned int i = 0; i < B->vertices.size(); i++)
  {
    Vector3 v = B->vertices[i];
    if (abs((v - plane_point_local_B)*normal_local_B) < depth+0.1)
    {
      verticesB.push_back(B->vertices[i]);
    }
  }
#else
  verticesA = A->vertices;
  verticesB = B->vertices;
#endif
  
  for (unsigned int i = 0; i < verticesA.size(); i++)
  {
    Vector3 global_p = A->RotationMatrix()*verticesA[i] + A->pos;
    if (IsPointInside(global_p, B))
    {
      CollisionPoint cp;
      cp.p  = verticesA[i];
      cp.eA = NULL;
      cp.eB = NULL;
      cp.f  = MinimumPenFaceSecond(A, B);
      cp.n = B->RotationMatrix()*cp.f->n;
      ret->points[0].push_back(cp);
      A->RememberColPoint(verticesA[i]);
    }
  }
  
  for (unsigned int i = 0; i < verticesB.size(); i++)
  {
    Vector3 global_p = B->RotationMatrix()*verticesB[i] + B->pos;
    if (IsPointInside(global_p, A))
    {
      CollisionPoint cp;
      cp.p  = verticesB[i];
      cp.eA = NULL;
      cp.eB = NULL;
      cp.f  = MinimumPenFaceSecond(B, A);
      cp.n = A->RotationMatrix()*cp.f->n;
      ret->points[1].push_back(cp);
      B->RememberColPoint(verticesB[i]);
    }
  }
  
  std::vector<Edge> edgesA;
  std::vector<Edge> edgesB;
  
#ifdef FILTER_EDGES
  //printf("normal: %f, %f, %f\n", normal_b_to_a.x(), normal_b_to_a.y(), normal_b_to_a.z());
  //printf("separating_point: %f, %f, %f\n", separating_plane_point.x(), separating_plane_point.y(), separating_plane_point.z());
  //printf("posA: %f, %f, %f\n", A->pos.x(), A->pos.y(), A->pos.z());
  //printf("posB: %f, %f, %f\n", B->pos.x(), B->pos.y(), B->pos.z());
  //printf("depth: %f\n", depth);
  
  for (unsigned int i = 0; i < A->edges.size(); i++)
  {
    Edge e = A->edges[i];
    if ((abs((e.a - plane_point_local_A)*normal_local_A) < depth+0.1) &&
        (abs((e.b - plane_point_local_A)*normal_local_A) < depth+0.1))
    {
      edgesA.push_back(A->edges[i]);
    }
  }
  for (unsigned int i = 0; i < B->edges.size(); i++)
  {
    Edge e = B->edges[i];
    if ((abs((e.a - plane_point_local_B)*normal_local_B) < depth+0.1) &&
        (abs((e.b - plane_point_local_B)*normal_local_B) < depth+0.1))
    {
      edgesB.push_back(B->edges[i]);
    }
  }
#else
  edgesA = A->edges;
  edgesB = B->edges;
#endif
  
  for (unsigned int i = 0; i < edgesA.size(); i++)
  {
    for (unsigned int j = 0; j < edgesB.size(); j++)
    {
      Edge ea = A->RotationMatrix()*(edgesA[i]) + A->pos;
      Edge eb = B->RotationMatrix()*(edgesB[j]) + B->pos;
      //Vector3 edge_cross = ea.dir^eb.dir;
      
      // edges parallel
      //if (edge_cross.Length() < 0.001) 
      //  continue;
    
      // normal not similar to separating axis
      //if (acos(edge_cross.Normalized()*normal_b_to_a) > PI/6 &&
      //    acos(edge_cross.Normalized()*normal_b_to_a*-1) > PI/6)
      //  continue;
        
      // one edge completely inside another object - we can take vertex instead
      //if ( (IsPointInside(ea.a, B) && IsPointInside(ea.b, B)) || 
      //     (IsPointInside(eb.a, A) && IsPointInside(eb.b, A)) )
      //  continue;
      
      Vector3 ca = ClosestOnEdgeA(ea, eb);
      Vector3 cb = ClosestOnEdgeA(eb, ea);
      
      // collision point is actually a vertex
      if (ca == ea.a || ca == ea.b || cb == eb.a || cb == eb.b) 
        continue;
      
      // pair to distant to be related to the separating axis
      if ((ca - cb).Length() > depth*2) continue;
      
      if (IsPointInside(ca, B) || IsPointInside(cb, A))
      {
        CollisionPoint cp;
        cp.p = A->PointGlobalToLocal(ca);
        A->RememberColPoint(cp.p);
        cp.eA = &(A->edges[i]);
        cp.eB = &(B->edges[j]);
        cp.f  = NULL;
        cp.n = normal_b_to_a;
        ret->points[0].push_back(cp);
      }
    }
  }

  return ret;
}

void CDetector::RememberPoints(Collision* col, RBody* A, RBody* B)
{
  for (unsigned int i = 0; i < col->points[0].size(); i++)
  {
    A->RememberColPoint(col->points[0][i].p);
  }
  for (unsigned int i = 0; i < col->points[1].size(); i++)
  {
    B->RememberColPoint(col->points[1][i].p);
  }
}

void CDetector::CheckAll(std::vector<RBody*>* bodies, 
                         std::vector<Collision*>* collisions,
                         std::vector<Collision*>* contacts,
                         double eps)
{
  for (unsigned int i = 0; i < collisions->size(); i++)
    delete (*collisions)[i];
  collisions->clear();
  
  for (unsigned int i = 0; i < contacts->size(); i++)
    delete (*contacts)[i];
  contacts->clear();
  
  for (unsigned int i = 0; i < bodies->size(); i++)
    for (unsigned int j = i+1; j < bodies->size(); j++)
    {
      if (DetectCollision((*bodies)[i], (*bodies)[j]))
      {
        Collision* col = GetCollisionPoints((*bodies)[i], (*bodies)[j]);
        if (col->points[0].size() == 0 && col->points[1].size() == 0) continue;

        if (fabs(HighestRelVel(col)) >= eps)
          collisions->push_back(col);
      }
    }
}
