#include "Matrix3x3.hpp"
#include "Vector3.hpp"

Matrix3x3::Matrix3x3(double a)
{
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      values[i][j] = a;
}

Matrix3x3::Matrix3x3()
{
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      values[i][j] = 0;
}

double& Matrix3x3::operator()(unsigned char i, unsigned char j)
{
  return values[i%3][j%3];
}

Matrix3x3 Matrix3x3::operator*(Matrix3x3 m)
{
  Matrix3x3 ret;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      ret(i,j) = values[i][0] * m(0,j) + values[i][1] * m(1,j) + values[i][2] * m(2,j);
  return ret;
}

Matrix3x3 Matrix3x3::operator+(Matrix3x3 m)
{
  Matrix3x3 ret;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      ret(i,j) = values[i][j] + m(i,j);
  return ret;
}

Matrix3x3 Matrix3x3::operator-(Matrix3x3 m)
{
  Matrix3x3 ret;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      ret(i,j) = values[i][j] - m(i,j);
  return ret;
}

Matrix3x3 Matrix3x3::operator*(double a)
{
  Matrix3x3 ret;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      ret(i,j) = values[i][j] * a;
  return ret;
}

Matrix3x3 Matrix3x3::operator/(double a)
{
  Matrix3x3 ret;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      ret(i,j) = values[i][j] / a;
  return ret;
}

Matrix3x3 Matrix3x3::operator=(Matrix3x3 m)
{
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      values[i][j] = m(i,j);
      
  return *this;
}

bool Matrix3x3::operator==(Matrix3x3 m)
{
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      if (values[i][j] != m(i,j))
        return false;
  return true;
}

Matrix3x3 Matrix3x3::ColumnCross(Vector3 v)
{
  Matrix3x3 ret;
  
  Vector3 a(values[0][0], values[1][0], values[2][0]);
  Vector3 b(values[0][1], values[1][1], values[2][1]);
  Vector3 c(values[0][2], values[1][2], values[2][2]);
  
  a = a^v;
  b = b^v;
  c = c^v;
  
  ret(0,0) = a[0];
  ret(1,0) = a[1];
  ret(2,0) = a[2];
  ret(0,1) = b[0];
  ret(1,1) = b[1];
  ret(2,1) = b[2];
  ret(0,2) = c[0];
  ret(1,2) = c[1];
  ret(2,2) = c[2];
  
  return ret;
}

Matrix3x3 Matrix3x3::Transpose()
{
  Matrix3x3 ret;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      ret(i,j) = values[j][i];
      
  return ret;
}

Matrix3x3 Matrix3x3::Inverse()
{
  Matrix3x3 ret;
  double det = 0;
  
  det += values[0][0] * (values[1][1]*values[2][2] - values[1][2]*values[2][1]);
  det -= values[0][1] * (values[2][2]*values[1][0] - values[1][2]*values[2][0]);
  det += values[0][2] * (values[1][0]*values[2][1] - values[1][1]*values[2][0]);
  
  ret(0,0) = values[1][1]*values[2][2] - values[1][2]*values[2][1];
  ret(0,1) = -(values[0][1]*values[2][2] - values[0][2]*values[2][1]);
  ret(0,2) = values[0][1]*values[1][2] - values[0][2]*values[1][1];
  ret(1,0) = -(values[1][0]*values[2][2] - values[1][2]*values[2][0]);
  ret(1,1) = values[0][0]*values[2][2] - values[0][2]*values[2][0];
  ret(1,2) = -(values[0][0]*values[1][2] - values[0][2]*values[1][0]);
  ret(2,0) = values[1][0]*values[2][1] - values[1][1]*values[2][0];
  ret(2,1) = -(values[0][0]*values[2][1] - values[0][1]*values[2][0]);
  ret(2,2) = (values[0][0]*values[1][1] - values[0][1]*values[1][0]);
  
  ret = ret*(1.0f/det);
  return ret;
}

Vector3 Matrix3x3::operator*(Vector3 v)
{
  Vector3 ret;
  
  for (int i = 0; i < 3; i++)
    ret[i] = values[i][0] * v[0] + values[i][1] * v[1] + values[i][2] * v[2];
  return ret;
}

Edge Matrix3x3::operator*(Edge e)
{
  Edge ret;
  ret.a = (*this)*e.a;
  ret.b = (*this)*e.b;
  ret.dir = (*this)*e.dir;
  
  return ret;
}

void Matrix3x3::Print()
{
  //for (int i = 0; i < 3; i++)
}
