#include "ODESolver.hpp"

#include <cstdio>

void ODESolver::EulerStepPos(RBody& rb, double delta_time)
{
  if (rb.is_static || rb.is_frozen) return;
    
  RBState derivative = rb.StateDerivative(rb.State());
  
  rb.pos = rb.pos + (derivative.pos * delta_time);
  rb.SetAngPos(rb.GetAngPos() + (derivative.ang_pos * delta_time));
}

void ODESolver::EulerStepVel(RBody& rb, double delta_time)
{  
  if (rb.is_static || rb.is_frozen) return;
  
  RBState derivative = rb.StateDerivative(rb.State());
  
  rb.vel = rb.vel + (derivative.vel * delta_time);
  rb.ang_vel = rb.ang_vel + (derivative.ang_vel * delta_time);
}

void ODESolver::RK4Step(RBody& rb, double delta_time)
{
  if (rb.is_static || rb.is_frozen) return;
  
  RBState k1 = rb.StateDerivative(rb.State());
  RBState k2 = rb.StateDerivative(rb.State() + k1*(delta_time/2.0f));
  RBState k3 = rb.StateDerivative(rb.State() + k2*(delta_time/2.0f));
  RBState k4 = rb.StateDerivative(rb.State() + k3*delta_time);
  RBState delta = (k1 + k2*2.0f + k3*2.0f + k4) * (1.0f/6.0f * delta_time);
  
  rb.pos = rb.pos + delta.pos;
  rb.vel = rb.vel + delta.vel;
  rb.SetAngPos(rb.GetAngPos() + delta.ang_pos);
  rb.ang_vel = rb.ang_vel + delta.ang_vel;
}

void ODESolver::RK4StepAll(std::vector<RBody*>* bodies, double delta_time)
{
  for (unsigned int i = 0; i < bodies->size(); i++)
    RK4Step(*(*bodies)[i], delta_time);
}

void ODESolver::EulerStepAll(std::vector<RBody*>* bodies, double delta_time)
{
  for (unsigned int i = 0; i < bodies->size(); i++)
  {
    EulerStepVel(*(*bodies)[i], delta_time);
    EulerStepPos(*(*bodies)[i], delta_time);
  }
}

void ODESolver::EulerStepAllVel(std::vector<RBody*>* bodies, double delta_time)
{
  for (unsigned int i = 0; i < bodies->size(); i++)
  {
    EulerStepVel(*(*bodies)[i], delta_time);
  }
}

void ODESolver::EulerStepAllPos(std::vector<RBody*>* bodies, double delta_time)
{
  for (unsigned int i = 0; i < bodies->size(); i++)
  {
    EulerStepPos(*(*bodies)[i], delta_time);
  }
}
