#include "Quaternion.hpp"
#include <cmath>
#include <cstdio>

Quaternion::Quaternion()
{
  s = 1;
  v = Vector3();
}

Quaternion::Quaternion(double a_s, Vector3 a_v)
{
  s = a_s;
  v = a_v;
}

Quaternion::Quaternion(double x, double y, double z, double w)
{
  s = x;
  v[0] = y;
  v[1] = z;
  v[2] = w;
}

Quaternion Quaternion::operator*(Quaternion q)
{
  Quaternion ret;
  ret.s = s*q.s - v*q.v;
  ret.v = v*q.s + q.v*s + (v^q.v);
  return ret;
}

Quaternion Quaternion::operator+(Quaternion q)
{
  Quaternion ret;
  ret.s = s + q.s;
  ret.v = v + q.v;
  return ret;
}

Quaternion Quaternion::operator*(double a)
{
  Quaternion ret;
  ret.s = s*a;
  ret.v = v*a;
  return ret;
}

Quaternion Quaternion::operator/(double a)
{
  Quaternion ret;
  ret.s = s/a;
  ret.v = v/a;
  return ret;
}

Quaternion Quaternion::operator=(Quaternion q)
{
  s = q.s;
  v = q.v;
  return *this;
}

Quaternion Quaternion::Normalize()
{
  double len = sqrt(s*s + v.x()*v.x() + v.y()*v.y() + v.z()*v.z());
  Quaternion ret;
  ret.s = s/len;
  ret.v = v/len;
  return ret;
}

void Quaternion::Print()
{
  printf("quaternion: %f, %f, %f, %f\n", s, v.x(), v.y(), v.z());
}
