#ifndef RBODY
#define RBODY

#include "Vector3.hpp"
#include "Matrix3x3.hpp"
#include "Quaternion.hpp"
#include "RBState.hpp"
#include "Face.hpp"
#include "Edge.hpp"
#include "Collision.hpp"
#include <vector>
#include <list>
#include <string>

#define COL_BUF_SIZE 4

class RBody
{
  Quaternion ang_pos;
  std::list<RBState> past_states;
    
  public:
    double density;
    double width, height, depth;
  
    Vector3 pos;
    Vector3 vel;
    Vector3 ang_vel;
    
    Vector3 forces;
    Vector3 torque;
    double  mass;
    
    bool is_static;
    bool is_frozen;
    
    Matrix3x3   local_inertia;
    Matrix3x3   inv_local_inertia;
    Matrix3x3   rot_matrix;
    Matrix3x3   inv_rot_matrix;
    bool        rot_matrix_valid;

    int         recent_cols_last;
    int         recent_cols_size;
    Vector3     recent_cols[COL_BUF_SIZE];
    Vector3     colour;
    std::string name;
    int         id;
    double      elasticity;
    double      bounding_radius;
    
    std::vector<Face>       faces;
    std::vector<Edge>       edges;
    std::vector<Vector3>    vertices;

    RBody(double dens, double x, double y, double z, std::string str = " ");
    
    Vector3 PointLocalToGlobal(Vector3 v);
    Vector3 VectorLocalToGlobal(Vector3 v);
    Vector3 PointGlobalToLocal(Vector3 v);
    Vector3 VectorGlobalToLocal(Vector3 v);
    
    Quaternion  GetAngPos() const;
    void        SetAngPos(Quaternion q);
    
    Vector3 Forces(RBState state) const;
    Vector3 Torque(RBState state) const;
    RBState StateDerivative(RBState s) const;
    RBState State() const;
    Face    GetFace(int i);
    void    ResetForceTorque();
    void    ApplyState(RBState state);
    void    ApplyPos(RBState state);
    void    ApplyVel(RBState state);
    void    RememberColPoint(Vector3 p);
    int     FaceNum();
    
    void PushState();
    void RevertState(int backward_count);
    
    Matrix3x3 RotationMatrix();
    Matrix3x3 InertiaTensor();
    Matrix3x3 InvInertiaTensor();
    
    static Matrix3x3 RotationMatrix(Quaternion q);
};

#endif
