#include "Renderer.hpp"
#include <GL/glut.h>
#include <math.h>

#include <cstdio>

namespace Renderer
{
  int width;
  int height;
  float lightPos[4] = { 10, 10, 10, 0 };
  Vector3 cameraPos(0, 50, 50);
  int font=(int)GLUT_BITMAP_8_BY_13;
  double camera_angle = 0;
  std::vector<RBody*>* bodies = NULL;
  std::vector<Collision*>* collisions = NULL;
  GLfloat rot_matrix[16];
  
  float camera_polar_angle = 1.5;
  float camera_longitude = 2;
  float camera_radius = 60;
}

void Renderer::InitDisplay(int* argc, char** argv, std::vector<RBody*>* b_ptr, std::vector<Collision*>* c_ptr, int w, int h)
{
  width = w;
  height = h;
  
  bodies = b_ptr;
  collisions = c_ptr;
  
  glutInit(argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_MULTISAMPLE);
  glutInitWindowSize(width, height);
  glutCreateWindow("Test");
  
  glEnable(GL_MULTISAMPLE);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  float ambient[4] = {0.5, 0.5, 0.5, 1};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient); 
  
  glutDisplayFunc(Display);
  glutReshapeFunc(Reshape);
  
  RotateCamera(0,0,0,0);
}

void Renderer::Reshape(int w, int h)
{
  width = w;
  height = h;

  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  gluPerspective(40.0, (float)width/(float)height, 0.5, 1000.0);
  glMatrixMode(GL_MODELVIEW);
}

void Renderer::DrawCollision(Collision* col)
{
  float color[4] = { 1, 0, 0, 1 };
  for (int body = 0; body < 2; body++)
  {
    for (unsigned int i = 0; i < col->points[body].size(); i++)
    {
      glPushMatrix();
	glTranslatef(col->points[body][i].p.x(), col->points[body][i].p.y(), col->points[body][i].p.z());
	glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
	glMaterialfv(GL_FRONT, GL_AMBIENT, color);
	glutSolidSphere(0.5f, 8, 8);
      glPopMatrix();
    }
  }
}

void Renderer::DrawBody(RBody* rb)
{
  double* matrix = new double[16];
  Matrix3x3 rb_matrix = rb->RotationMatrix();
  for (int i = 0; i < 16; i++) matrix[i] = 0;
  matrix[15] = 1;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      matrix[i+4*j] = rb_matrix(i,j);

  glPushMatrix();      
    glTranslatef(rb->pos[0], rb->pos[1], rb->pos[2]);
    glMultMatrixd(matrix);
    float color[4] = { (float)rb->colour[0], (float)rb->colour[1], (float)rb->colour[2], 1 };
    float color_edge[4] = { 0.0f, 0.0f, 0.0f, 1 };
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
    glMaterialfv(GL_FRONT, GL_AMBIENT, color);
    
    for (int i = 0; i < rb->FaceNum(); i++)
    {
      Face face = rb->GetFace(i);
      glBegin(GL_POLYGON);
	glNormal3f(face.n.x(), face.n.y(), face.n.z());
	for (unsigned int j = 0; j < face.vertices.size(); j++)
	  glVertex3f(face.vertices[j].x(), face.vertices[j].y(), face.vertices[j].z());
      glEnd();
    }
    
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color_edge);
    glMaterialfv(GL_FRONT, GL_AMBIENT, color_edge);
    
    for (unsigned int i = 0; i < rb->edges.size(); i++)
    {
      Edge edge = rb->edges[i];
      glLineWidth(2.5);
      glBegin(GL_LINES);
	glVertex3f(edge.a.x(), edge.a.y(), edge.a.z());
	glVertex3f(edge.b.x(), edge.b.y(), edge.b.z());
      glEnd();
    }
    
  glPopMatrix();
}

void Renderer::DrawFloor(double x, double y, double z, double rows, double cols)
{
  float colorWhite[4] = { 1, 1, 1, 1 };
  float colorGrey[4] = { 0.5, 0.5, 0.5, 1 };
  
  glPushMatrix();
    glTranslatef(x, y, z);
    for (int i = 0; i < rows; i++)
      for (int j = 0; j < cols; j++)
      {
	glPushMatrix();
	  glTranslatef(i, j, 0);
	  if ((i+j)%2 == 0)
	  {
	    glMaterialfv(GL_FRONT, GL_DIFFUSE, colorWhite);
	    glMaterialfv(GL_FRONT, GL_AMBIENT, colorWhite);
	  }
	  else
	  {
	    glMaterialfv(GL_FRONT, GL_DIFFUSE, colorGrey);
	    glMaterialfv(GL_FRONT, GL_AMBIENT, colorGrey);
	  }
	  glBegin(GL_QUADS);
	    glVertex3f(1, 1, 0);
	    glVertex3f(1, 0, 0);
	    glVertex3f(0, 0, 0);
	    glVertex3f(0, 1, 0);
	  glEnd();
	glPopMatrix();
      }    
  glPopMatrix();
}

void Renderer::RenderBitmapString(float x, float y, void *font, std::string str)
{  
  glRasterPos2f(x, y);
  
  for (unsigned int i = 0; i < str.size(); i++) {
    glutBitmapCharacter(font, str[i]);
  }
}

void Renderer::Display()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
  
  gluLookAt(cameraPos[0], cameraPos[1], cameraPos[2], 
    0.0, 0.0, 0.0, 0.0, 0.0, 1.0);

  if (bodies != NULL && collisions != NULL)
  {
    for (unsigned int i = 0; i < bodies->size(); i++)
    {
      DrawBody((*bodies)[i]);
    }
    
    for (unsigned int i = 0; i < collisions->size(); i++)
    {
      DrawCollision((*collisions)[i]);
    }
  }
  //renderBitmapString(20,35,(void *)font, "fps: "+std::to_string(fps));
  //renderBitmapString(20,30,(void *)font, "col: "+std::to_string(are_colliding));
  
  glutSwapBuffers();
}

void Renderer::SetCamera(double angle)
{
  camera_angle = angle;
}

int Renderer::GetWidth()
{
  return width;
}

int Renderer::GetHeight()
{
  return height;
}

void Renderer::RotateCamera(float v, float h, bool up, bool down)
{
  if (up) camera_radius -= 4;
  if (down) camera_radius += 4;
  
  if (camera_radius < 5)
    camera_radius = 5;
  
  camera_polar_angle -= v/200;
  if (camera_polar_angle > 3.1) camera_polar_angle = 3.1;
  if (camera_polar_angle < 0.1) camera_polar_angle = 0.1;
  camera_longitude -= h/200;
  
  cameraPos[0] = camera_radius * sin(camera_polar_angle) * cos(camera_longitude);
  cameraPos[1] = camera_radius * sin(camera_polar_angle) * sin(camera_longitude);
  cameraPos[2] = camera_radius * cos(camera_polar_angle);
}
