#include "Vector3.hpp"
#include <cmath>
#include <cstdio>

#define COMPARE_EPS 0.1

Vector3::Vector3(double x, double y, double z)
{
  axes[0] = x;
  axes[1] = y;
  axes[2] = z;
  len_correct = false;
}

Vector3::Vector3()
{
  axes[0] = 0;
  axes[1] = 0;
  axes[2] = 0;
  len_correct = false;
}

double& Vector3::operator[](unsigned char i)
{
  len_correct = false;
  return axes[i];
}

double Vector3::Length()
{
  if (!len_correct)
  {
    len = sqrt(axes[0]*axes[0] + axes[1]*axes[1] + axes[2]*axes[2]);
    len_correct = true;
  }
  return len;
}

Vector3 Vector3::operator*(double a)
{
  Vector3 ret(axes[0] * a, axes[1] * a, axes[2] * a);
  return ret;
}

Vector3 Vector3::operator/(double a)
{
  if (a == 0) return Vector3();
  
  Vector3 ret(axes[0] / a, axes[1] / a, axes[2] / a);
  return ret;
}

double Vector3::operator*(Vector3 v)
{
  double ret = 0;
  ret += axes[0] * v.x();
  ret += axes[1] * v.y();
  ret += axes[2] * v.z();
  return ret;
}

Vector3 Vector3::operator+(Vector3 v)
{
  Vector3 ret(axes[0] + v.x(), axes[1] + v.y(), axes[2] + v.z());
  return ret;
}

Vector3 Vector3::operator-(Vector3 v)
{
  Vector3 ret(axes[0] - v.x(), axes[1] - v.y(), axes[2] - v.z());
  return ret;
}

Vector3 Vector3::operator=(Vector3 v)
{
  axes[0] = v.x();
  axes[1] = v.y();
  axes[2] = v.z();
  return *this;
}

bool Vector3::operator==(Vector3 v)
{
  if (axes[0] == v.x() && axes[1] == v.y() && axes[2] == v.z())
    return true;
  else
    return false;
}

Vector3 Vector3::operator^(Vector3 v)
{
  Vector3 ret;
  ret[0] = axes[1]*v[2] - axes[2]*v[1];
  ret[1] = axes[2]*v[0] - axes[0]*v[2];
  ret[2] = axes[0]*v[1] - axes[1]*v[0];
  return ret;
}

Vector3 Vector3::Normalized()
{
  Vector3 ret = *this;
  if (ret.Length() != 0)
    ret = ret/ret.Length();
  return ret;
}

Vector3 Vector3::operator+=(Vector3 v)
{
  axes[0] += v.x();
  axes[1] += v.y();
  axes[2] += v.z();
  return *this;
}

Vector3 Vector3::operator-=(Vector3 v)
{
  axes[0] -= v.x();
  axes[1] -= v.y();
  axes[2] -= v.z();
  return *this;
}

bool Vector3::IsAlmost(Vector3 v)
{
  if (axes[0] < v.x()-COMPARE_EPS || axes[0] > v.x()+COMPARE_EPS) return false; 
  if (axes[1] < v.y()-COMPARE_EPS || axes[1] > v.y()+COMPARE_EPS) return false; 
  if (axes[2] < v.z()-COMPARE_EPS || axes[2] > v.z()+COMPARE_EPS) return false; 
  else return true;
}

void Vector3::Print()
{
  printf("vector: %f, %f, %f\n", x(), y(), z());
}
